@@ -1,22 +1,46 @@
 """Main Controller"""
-
-from tg import expose, redirect, config, validate, override_template, response
-from tg.decorators import paginate, use_custom_format, with_trailing_slash
-from tg.render import render
+import tg
+from tg import expose, redirect, config, validate, override_template, response, render_template, tmpl_context
+from tg import cache, i18n, request
+from tg.decorators import paginate, use_custom_format, with_trailing_slash, Decoration, before_render, decode_params
 from tg.controllers import TGController
-from tw.forms import TableForm, TextField, CalendarDatePicker, SingleSelectField, TextArea
-from tw.api import WidgetsList
-from formencode import validators
+from tg.validation import TGValidationError
+from tg._compat import PY3
+from tg.render import _get_tg_vars, cached_template
+import datetime
 
-class MovieForm(TableForm):
-    # This WidgetsList is just a container
-    class fields(WidgetsList):
-        title = TextField()
-        year = TextField(size=4, default=1984)
-        description = TextArea()
+if not PY3:
+    from tw.forms import TableForm, TextField, CalendarDatePicker, SingleSelectField, TextArea
+    from tw.api import WidgetsList
 
-#then, we create an instance of this form
-base_movie_form = MovieForm("movie_form", action='create')
+    class MovieForm(TableForm):
+        # This WidgetsList is just a container
+        class fields(WidgetsList):
+            title = TextField()
+            year = TextField(size=4, default=1984)
+            description = TextArea()
+
+    #then, we create an instance of this form
+    base_movie_form = MovieForm("movie_form", action='create')
+
+
+    import tw2.forms as tw2f
+    import tw2.core as tw2c
+    class TW2MovieForm(tw2f.TableForm):
+        title = tw2f.TextField(validator=tw2c.Required)
+        year = tw2f.TextField(size=4, validator=tw2c.IntValidator)
+
+    tw2_movie_form = TW2MovieForm()
+
+else:
+    base_movie_form = None
+
+class IntValidator(object):
+    def to_python(self, value):
+        try:
+            return int(value)
+        except:
+            raise TGValidationError('Not a number')
 
 
 class GoodJsonObject(object):
@@ -52,9 +76,45 @@ class JsonController(TGController):
         return dict(obj=BadJsonObject())
 
 
+class SubClassableController(TGController):
+    @expose('genshi:index.html')
+    def index(self):
+        return {}
+
+    @expose('genshi:index.html')
+    def index_override(self):
+        return {}
+
+    def before_render_data(remainder, params, output):
+        output['parent_value'] = 'PARENT'
+
+    @expose('json')
+    @before_render(before_render_data)
+    def data(self):
+        return {'v':5}
+
+class SubClassingController(SubClassableController):
+    @expose(inherit=True)
+    def index(self, *args, **kw):
+        return super(SubClassingController, self).index(*args, **kw)
+
+    @expose('genshi:genshi_doctype.html', inherit=True)
+    def index_override(self, *args, **kw):
+        return super(SubClassingController, self).index_override(*args, **kw)
+
+    def before_render_data(remainder, params, output):
+        output['child_value'] = 'CHILD'
+
+    @expose(inherit=True)
+    @before_render(before_render_data)
+    def data(self, *args, **kw):
+        return super(SubClassingController, self).data(*args, **kw)
+
 class RootController(TGController):
 
     j = JsonController()
+    sub1 = SubClassableController()
+    sub2 = SubClassingController()
 
     @expose('genshi:index.html')
     def index(self):
@@ -88,6 +148,10 @@ class RootController(TGController):
     def form(self):
         return dict(form=base_movie_form)
 
+    @expose('genshi:genshi_form.html')
+    def tw2form(self):
+        return dict(form=tw2_movie_form)
+
     @expose('genshi:genshi_foreign.html')
     def foreign(self):
         return {}
@@ -96,11 +160,17 @@ class RootController(TGController):
     @validate(form=base_movie_form)
     def process_form_errors(self, **kwargs):
         #add error messages to the kwargs dictionary and return it
-        kwargs['errors'] = pylons.tmpl_context.form_errors
+        kwargs['errors'] = request.validation['errors']
         return dict(kwargs)
 
-    @expose('genshi:genshi_paginated.html')
+    @expose()
     @paginate('testdata')
+    def paginated_text(self):
+        return '''Some Text'''
+
+    @expose('genshi:genshi_paginated.html')
+    @expose('json')
+    @paginate('testdata', max_items_per_page=20)
     def paginated(self, n):
         return dict(testdata=range(int(n)))
 
@@ -112,12 +182,12 @@ class RootController(TGController):
 
     @expose('genshi:genshi_paginated.html')
     @paginate('testdata')
-    @validate(dict(n=validators.Int()))
+    @validate(dict(n=IntValidator()))
     def paginated_validated(self, n):
         return dict(testdata=range(n))
 
     @expose('genshi:genshi_paginated.html')
-    @validate(dict(n=validators.Int()))
+    @validate(dict(n=IntValidator()))
     @paginate('testdata')
     def validated_paginated(self, n):
         return dict(testdata=range(n))
@@ -141,11 +211,15 @@ class RootController(TGController):
     def genshi_inherits_sub_from_bottom(self):
         return {}
 
-    @expose('jinja:jinja_noop.html')
+    @expose('jinja:jinja_noop.jinja')
     def jinja_index(self):
         return {}
 
-    @expose('jinja:jinja_inherits.html')
+    @expose('jinja:jinja_autoload.jinja')
+    def jinja_autoload(self):
+        return {}
+
+    @expose('jinja:jinja_inherits.jinja')
     def jinja_inherits(self):
         return {}
 
@@ -161,21 +235,35 @@ class RootController(TGController):
     def jinja_inherits_mixed(self):
         return {}
 
-    @expose('jinja:jinja_extensions.html')
+    @expose('jinja:jinja_extensions.jinja')
     def jinja_extensions(self):
         test_autoescape_on = "<b>Test Autoescape On</b>"
         test_autoescape_off = "<b>Autoescape Off</b>"
         return dict(test_autoescape_off=test_autoescape_off,
                 test_autoescape_on=test_autoescape_on)
 
-    @expose('jinja:jinja_filters.html')
+    @expose('jinja:jinja_filters.jinja')
     def jinja_filters(self):
         return {}
 
-    @expose('jinja:jinja_buildins.html')
+    @expose('jinja:jinja_buildins.jinja')
     def jinja_buildins(self):
         return {}
 
+    @expose('jinja:jinja_i18n.jinja')
+    def jinja_i18n(self):
+        return {}
+
+    @expose('jinja:jinja_i18n.jinja')
+    def jinja_i18n_en(self):
+        i18n.set_temporary_lang("en")
+        return {}
+
+    @expose('jinja:jinja_i18n.jinja')
+    def jinja_i18n_de(self):
+        i18n.set_temporary_lang("de")
+        return {}
+
     @expose('chameleon_genshi:index.html')
     def chameleon_genshi_index(self):
         return {}
@@ -269,15 +357,102 @@ class RootController(TGController):
 
     @expose()
     def jinja2_manual_rendering(self, frompylons=False):
+        try:
+            import pylons
+        except ImportError:
+            frompylons = False
+
         if frompylons:
             from pylons.templating import render_jinja2
-            return render_jinja2('jinja_inherits.html')
+            return render_jinja2('jinja_inherits.jinja')
         else:
-            return render({}, 'jinja', 'jinja_inherits.html')
+            return render_template({}, 'jinja', 'jinja_inherits.jinja')
 
+    @expose()
+    def no_template_generator(self):
+        def output():
+            num = 0
+            while num < 5:
+                num += 1
+                yield str(num).encode('ascii')
+        return output()
+    
     @expose()
     def genshi_manual_rendering_with_doctype(self, doctype=None):
         response.content_type = 'text/html'
         response.charset = 'utf-8'
-        return render({}, 'genshi', 'genshi_doctype.html', doctype=doctype)
+        return render_template({}, 'genshi', 'genshi_doctype.html', doctype=doctype)
 
+    @expose('mako:mako_custom_format.mak')
+    @expose('genshi:genshi_custom_format.html')
+    def multiple_engines(self):
+        deco = Decoration.get_decoration(self.multiple_engines)
+        used_engine = deco.engines.get('text/html')[0]
+        return dict(format=used_engine, status='ok')
+
+    @expose('json')
+    def get_tg_vars(self):
+        return dict(tg_vars=list(_get_tg_vars().keys()))
+
+    @expose('genshi:index.html')
+    def template_caching(self):
+        from datetime import datetime
+        tmpl_context.now = datetime.utcnow
+        return dict(tg_cache={'key':'TEMPLATE_CACHE_TEST',
+                              'type':'memory',
+                              'expire':'never'})
+
+    @expose('genshi:index.html')
+    def template_caching_default_type(self):
+        from datetime import datetime
+        tmpl_context.now = datetime.utcnow
+        return dict(tg_cache={'key':'TEMPLATE_CACHE_TEST2',
+                              'expire':'never'})
+
+    @expose('json')
+    def template_caching_options(self, **kwargs):
+        _cache_options = {}
+        class FakeCache(object):
+            def get_cache(self, *args, **kwargs):
+                _cache_options['args'] = args
+                _cache_options['kwargs'] = kwargs
+                try:
+                    c = cache.get_cache(*args, **kwargs)
+                    _cache_options['cls'] = c.namespace.__class__.__name__
+                except TypeError:
+                    _cache_options['cls'] = 'NoImplementation'
+                    c = cache.get_cache(*args, type='memory', **kwargs)
+                return c
+
+        tg.cache.kwargs['type'] = 'NoImplementation'
+        old_cache = tg.cache
+        tg.cache = FakeCache()
+
+        try:
+            def render_func(*args, **kw):
+                return 'OK'
+            cached_template('index.html', render_func, **kwargs)
+            return _cache_options
+        finally:
+            tg.cache = old_cache
+
+    @expose('jsonp', render_params={'callback_param': 'call'})
+    def get_jsonp(self, **kwargs):
+        return {'value': 5}
+
+    @expose('json')
+    def get_json_isodates_default(self, **kwargs):
+        return {'date': datetime.datetime.utcnow()}
+
+    @expose('json', render_params={'isodates': True})
+    def get_json_isodates_on(self, **kwargs):
+        return {'date': datetime.datetime.utcnow()}
+
+    @expose('json', render_params={'isodates': False})
+    def get_json_isodates_off(self, **kwargs):
+        return {'date': datetime.datetime.utcnow()}
+
+    @expose('json')
+    @decode_params('json')
+    def echo_json(self, **kwargs):
+        return kwargs
