@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 """
-repoze.what & repoze.what-pylons **integration** tests.
+repoze.who **integration** tests.
 
 Note that it is not necessary to have integration tests for the other auth*
 software in this package. They must be in tg.devtools, specifically in the test
@@ -11,28 +11,28 @@ this writing).
 
 from unittest import TestCase
 from shutil import rmtree
-from urllib import unquote
 import os
+from nose.tools import raises
 
-from paste import httpexceptions
-from paste.registry import RegistryManager
+from tg._compat import url_unquote
+from tg.configuration.utils import TGConfigError
+
+from tg.support.registry import RegistryManager
 from webob import Response, Request
 from webtest import TestApp
-from beaker.middleware import CacheMiddleware, SessionMiddleware
 
-from tg import request, response, expose, require, allow_only
+from tg import request, response, expose, require, redirect
 from tg.controllers import TGController, WSGIAppController, RestController
-import pylons
-from pylons import tmpl_context
-from pylons.controllers.util import abort
-from pylons.util import ContextObj, PylonsContext
-from pylons.testutil import ControllerWrap, SetupCacheGlobal
+from tg.controllers.util import abort, auth_force_login, auth_force_logout
+from tg.wsgiapp import TGApp
+from tg.support.middlewares import CacheMiddleware, SessionMiddleware, StatusCodeRedirect
+from tg.decorators import Decoration
 
-from repoze.who.plugins.auth_tkt import AuthTktCookiePlugin
-from repoze.what.middleware import setup_auth
-from repoze.what.predicates import Not, is_user, not_anonymous
+from .baseutils import ControllerWrap, FakeRoutes, default_config
+
+from tg.configuration.auth import setup_auth, TGAuthMetadata
+from tg.predicates import is_user, not_anonymous, in_group, has_permission
 
-from pylons.middleware import StatusCodeRedirect
 from tg.error import ErrorHandler
 
 #{ AUT's setup
@@ -44,11 +44,49 @@ session_dir = os.path.join(data_dir, 'session')
 # Just in case...
 rmtree(session_dir, ignore_errors=True)
 
-def make_app(controller_klass, environ={}, with_errors=False):
+
+class TestAuthMetadata(TGAuthMetadata):
+    """
+    Provides a way to lookup for user, groups and permissions
+    given the current identity. This has to be specialized
+    for each storage backend.
+
+    By default it returns empty lists for groups and permissions
+    and None for the user.
+    """
+    def get_user(self, identity, userid):
+        if ':' in userid:
+            return userid.split(':')[0]
+
+        return super(TestAuthMetadata, self).get_user(identity, userid)
+
+    def get_groups(self, identity, userid):
+        if userid:
+            parts = userid.split(':')
+            return parts[1:2]
+
+        return super(TestAuthMetadata, self).get_groups(identity, userid)
+
+    def get_permissions(self, identity, userid):
+        if userid:
+            parts = userid.split(':')
+            return parts[2:]
+
+        return super(TestAuthMetadata, self).get_permissions(identity, userid)
+
+
+def make_app(controller_klass, environ={}, with_errors=False, config_options=None):
     """Creates a ``TestApp`` instance."""
     # The basic middleware:
-    app = ControllerWrap(controller_klass)
-    app = SetupCacheGlobal(app, environ, setup_cache=True, setup_session=True)
+    conf = default_config.copy()
+    if config_options is not None:
+        conf.update(config_options)
+
+    app = TGApp(config=conf)
+    app.controller_classes['root'] = ControllerWrap(controller_klass)
+
+    app = FakeRoutes(app)
+    
     if with_errors:
         app = ErrorHandler(app, {}, debug=False)
         app = StatusCodeRedirect(app, [403, 404, 500])
@@ -56,19 +94,19 @@ def make_app(controller_klass, environ={}, with_errors=False):
     app = SessionMiddleware(app, {}, data_dir=session_dir)
     app = CacheMiddleware(app, {}, data_dir=os.path.join(data_dir, 'cache'))
 
-    # We're not going to use groups or permissions-related predicates here:
-    groups_adapters = None
-    permissions_adapters = None
-
     # Setting repoze.who up:
+    from repoze.who.plugins.auth_tkt import AuthTktCookiePlugin
     cookie = AuthTktCookiePlugin('secret', 'authtkt')
     identifiers = [('cookie', cookie)]
 
-    app = setup_auth(app, groups_adapters, permissions_adapters,
-                     identifiers=identifiers, authenticators=[],
-                     challengers=[], skip_authentication=True)
+    app = setup_auth(app, TestAuthMetadata(),
+                     identifiers=identifiers, skip_authentication=True,
+                     authenticators=[], challengers=[])
+
+    # As setup_auth with skip_authentication sets empty authenticators always
+    # we must manually append it after creating the middleware.
+    app.api_factory.authenticators.append(('cookie', cookie))
 
-    app = httpexceptions.make_middleware(app)
     return TestApp(app)
 
 
@@ -121,8 +159,28 @@ class ControlPanel(TGController):
     def add_user(self, user_name):
         return "%s was just registered" % user_name
 
-class RootController(TGController):
+class CustomAllowOnly(TGController):
+    class something(object):
+        def check_authorization(self, env):
+            from tg.controllers.decoratedcontroller import NotAuthorizedError
+            raise NotAuthorizedError()
 
+    @expose()
+    def index(self):
+        return 'HI'
+
+    allow_only = something()
+
+class SmartDenialAllowOnly(TGController):
+    allow_only = require(is_user('developer'), smart_denial=True)
+
+    @expose('json')
+    def data(self):
+        return {'key': 'value'}
+
+class RootController(TGController):
+    custom_allow = CustomAllowOnly()
+    smart_allow = SmartDenialAllowOnly()
     cp = ControlPanel()
 
     rest = DaRestController()
@@ -138,6 +196,28 @@ class RootController(TGController):
     def commit(self):
         return 'you can commit'
 
+    @expose('json:')
+    @require(is_user('developer'), smart_denial=True)
+    def smartabort(self):
+        return {'key': 'value'}
+
+    @expose()
+    @require(in_group('managers'))
+    @require(has_permission('commit'))
+    def force_commit(self):
+        return 'you can commit'
+
+    @expose()
+    def login_logout(self, username, noidentifier='0'):
+        if noidentifier == '1':
+            request.environ['repoze.who.plugins'] = {}
+
+        if username == 'logout':
+            auth_force_logout()
+        else:
+            auth_force_login('%s:managers' % username)
+
+        return 'OK'
 
 class ControllerWithAllowOnlyAttributeAndAuthzDenialHandler(TGController):
     """Mock TG2 protected controller using the .allow_only attribute"""
@@ -154,26 +234,6 @@ class ControllerWithAllowOnlyAttributeAndAuthzDenialHandler(TGController):
         abort(402)
 
 
-class ControllerWithAllowOnlyDecoratorAndAuthzDenialHandler(TGController):
-    """
-    Mock TG2 protected controller using the @allow_only decorator, but also
-    using ._failed_authorization()
-
-    """
-
-    @expose()
-    def index(self):
-        return 'Welcome back, foobar!'
-
-    @classmethod
-    def _failed_authorization(self, reason):
-        # Pay first!
-        abort(402)
-
-ControllerWithAllowOnlyDecoratorAndAuthzDenialHandler = allow_only(
-    is_user('foobar'))(ControllerWithAllowOnlyDecoratorAndAuthzDenialHandler)
-
-
 #{ The tests themselves
 
 
@@ -186,19 +246,11 @@ class BaseIntegrationTests(TestCase):
         # Creating the session dir:
         if not os.path.exists(session_dir):
             os.makedirs(session_dir)
+
         # Setting TG2 up:
-        c = ContextObj()
-        py_obj = PylonsContext()
-        py_obj.c = c
-        py_obj.request = py_obj.response = None
-        environ = {'pylons.routes_dict': dict(action='index'),
-                   'pylons.pylons': py_obj}
-        pylons.tmpl_context._push_object(c)
-        # Finally, the app:
-        self.app = make_app(self.controller, environ)
+        self.app = make_app(self.controller, {}, config_options=getattr(self, 'CONFIG_OPTIONS', {}))
 
     def tearDown(self):
-        tmpl_context._pop_object()
         # Removing the session dir:
         rmtree(session_dir, ignore_errors=True)
 
@@ -208,49 +260,145 @@ class BaseIntegrationTests(TestCase):
 
         """
         assert 'webflash' in response.cookies_set, "Such no WebFlash cookie"
-        flash = unquote(response.cookies_set['webflash'])
+        flash = url_unquote(response.cookies_set['webflash'])
         for msg in expected_messages:
             msg = '"%s"' % msg
             assert msg in flash, 'Message %s not in flash: %s' % (msg, flash)
 
+class TestLoginLogout(BaseIntegrationTests):
+    def test_user_login(self):
+        resp = self.app.get('/login_logout?username=developer')
+        cookie = resp.headers['Set-Cookie']
+        assert 'authtkt="' in cookie
+        assert 'developer' in cookie
+
+    @raises(TGConfigError)
+    def test_user_login_without_identifier(self):
+        resp = self.app.get('/login_logout?username=developer&noidentifier=1')
+
+    def test_user_logout(self):
+        AUTH_TOKEN = 'authtkt="f0657f514a6b960d50ea199aea76534d53cd0dd4developer%3Amanagers!"'
+        resp = self.app.get('/login_logout?username=logout', headers=[('Cookie', AUTH_TOKEN)])
+        assert 'authtkt="INVALID"' in str(resp), resp
+
+    def test_user_override(self):
+        AUTH_TOKEN = 'authtkt="f0657f514a6b960d50ea199aea76534d53cd0dd4developer%3Amanagers!"'
+        resp = self.app.get('/login_logout?username=admin', headers=[('Cookie', AUTH_TOKEN)])
+        cookie = resp.headers['Set-Cookie']
+        assert 'authtkt="' in cookie, resp
+        assert 'admin' in cookie, resp
 
 class TestRequire(BaseIntegrationTests):
     """Test case for the @require decorator"""
 
+    def test_authz_custom_allow_only(self):
+        #environ = {'REMOTE_USER': 'developer'}
+        resp = self.app.get('/custom_allow', extra_environ={}, status=401)
+
     def test_authz_granted_in_root_controller(self):
         environ = {'REMOTE_USER': 'developer'}
         resp = self.app.get('/commit', extra_environ=environ, status=200)
-        self.assertEqual("you can commit", resp.body)
+        self.assertEqual("you can commit", resp.body.decode('utf-8'))
+
+    def test_multiple_requirements_passed(self):
+        environ = {'REMOTE_USER': 'developer:managers:commit'}
+        resp = self.app.get('/force_commit', extra_environ=environ, status=200)
+        self.assertEqual("you can commit", resp.text)
+
+    def test_multiple_requirements_blocked_1(self):
+        environ = {'REMOTE_USER': 'tester:testing:commit'}
+        resp = self.app.get('/force_commit', extra_environ=environ, status=403)
+        assert 'The current user must belong to the group "managers"' in resp.text, resp.text
+
+    def test_multiple_requirements_blocked_2(self):
+        environ = {'REMOTE_USER': 'manager:managers:viewonly'}
+        resp = self.app.get('/force_commit', extra_environ=environ, status=403)
+        assert 'The user must have the "commit" permission' in resp.text, resp.text
+
+    def test_multiple_requirements_all_registered(self):
+        deco = Decoration.get_decoration(RootController.force_commit)
+        assert len(deco.requirements) == 2, deco.requirements
+
+    def test_multiple_requirements_backward_compatibility(self):
+        deco = Decoration.get_decoration(RootController.force_commit)
+        predicate = deco.requirement.predicate
+        assert isinstance(predicate, has_permission), predicate
+
+    def test_no_requirements_backward_compatibility(self):
+        deco = Decoration.get_decoration(RootController.force_commit)
+        reqs = deco.requirements
+        deco.requirements = []
+        requirement = deco.requirement
+        deco.requirements = reqs
+        assert requirement is None, requirement
 
     def test_authz_denied_in_root_controller(self):
         # As an anonymous user:
         resp = self.app.get('/commit', status=401)
-        assert "you can commit" not in resp.body
+        assert "you can commit" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"developer\"')
         # As an authenticated user:
         environ = {'REMOTE_USER': 'foobar'}
         resp = self.app.get('/commit', extra_environ=environ, status=403)
-        assert "you can commit" not in resp.body
+        assert "you can commit" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"developer\"')
 
     def test_authz_granted_in_sub_controller(self):
         environ = {'REMOTE_USER': 'admin'}
         resp = self.app.get('/cp/add_user/foo', extra_environ=environ,
                             status=200)
-        self.assertEqual("foo was just registered", resp.body)
+        self.assertEqual("foo was just registered", resp.body.decode('utf-8'))
 
     def test_authz_denied_in_sub_controller(self):
         # As an anonymous user:
         resp = self.app.get('/cp/add_user/foo', status=401)
-        assert "was just registered" not in resp.body
+        assert "was just registered" not in resp.body.decode('utf-8')
         self._check_flash(resp, NOT_AUTHENTICATED)
         # As an authenticated user:
         environ = {'REMOTE_USER': 'foobar'}
         resp = self.app.get('/cp/add_user/foo', extra_environ=environ,
                             status=403)
-        assert "was just registered" not in resp.body
+        assert "was just registered" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"admin\"')
 
+    def test_smart_auth_json(self):
+        nouser = {'accept': 'application/json'}
+        baduser = {'accept': 'application/json',
+                'REMOTE_USER': 'foobar'}
+        gooduser = {'accept': 'application/json',
+                'REMOTE_USER': 'developer'}
+
+        resp = self.app.get('/smartabort.json', extra_environ=nouser, status=401)
+        assert resp.status == '401 Unauthorized', 'Expected 401, got %s' % (resp.status)
+        assert 'The current user must be "developer"' in resp.json['detail']
+
+        resp = self.app.get('/smartabort.json', extra_environ=baduser, status=403)
+        assert resp.status == '403 Forbidden', 'Expected 403, got %s' % (resp.status)
+        assert 'The current user must be "developer"' in resp.json['detail']
+
+        resp = self.app.get('/smartabort.json', extra_environ=gooduser, status=200)
+        assert resp.status == '200 OK', 'Expected 200, got %s' % (resp.body)
+        assert {'key': 'value'} == resp.json, resp.json
+
+    def test_smart_auth_json_allow_only(self):
+        nouser = {'accept': 'application/json'}
+        baduser = {'accept': 'application/json',
+                'REMOTE_USER': 'foobar'}
+        gooduser = {'accept': 'application/json',
+                'REMOTE_USER': 'developer'}
+
+        resp = self.app.get('/smart_allow/data.json', extra_environ=nouser, status=401)
+        assert resp.status == '401 Unauthorized', 'Expected 401, got %s' % (resp.status)
+        assert 'The current user must be "developer"' in resp.json['detail']
+
+        resp = self.app.get('/smart_allow/data.json', extra_environ=baduser, status=403)
+        assert resp.status == '403 Forbidden', 'Expected 403, got %s' % (resp.status)
+        assert 'The current user must be "developer"' in resp.json['detail']
+
+        resp = self.app.get('/smart_allow/data.json', extra_environ=gooduser, status=200)
+        assert resp.status == '200 OK', 'Expected 200, got %s' % (resp.body)
+        assert {'key': 'value'} == resp.json, resp.json
+
 
 class TestAllowOnlyDecoratorInSubController(BaseIntegrationTests):
     """Test case for the @allow_only decorator in a sub-controller"""
@@ -258,44 +406,24 @@ class TestAllowOnlyDecoratorInSubController(BaseIntegrationTests):
     def test_authz_granted_without_require(self):
         environ = {'REMOTE_USER': 'someone'}
         resp = self.app.get('/cp/', extra_environ=environ, status=200)
-        self.assertEqual("you are in the panel", resp.body)
+        self.assertEqual("you are in the panel", resp.body.decode('utf-8'))
 
     def test_authz_denied_without_require(self):
         resp = self.app.get('/cp/', status=401)
-        assert "you are in the panel" not in resp.body
+        assert "you are in the panel" not in resp.body.decode('utf-8')
         self._check_flash(resp, NOT_AUTHENTICATED)
 
     def test_authz_granted_with_require(self):
         environ = {'REMOTE_USER': 'admin'}
         resp = self.app.get('/cp/add_user/foo', extra_environ=environ,
                             status=200)
-        self.assertEqual("foo was just registered", resp.body)
+        self.assertEqual("foo was just registered", resp.body.decode('utf-8'))
 
     def test_authz_denied_with_require(self):
         resp = self.app.get('/cp/add_user/foo', status=401)
-        assert "was just registered" not in resp.body
+        assert "was just registered" not in resp.body.decode('utf-8')
         self._check_flash(resp, NOT_AUTHENTICATED)
 
-
-class _TestAllowOnlyDecoratorAndDefaultAuthzDenialHandler(BaseIntegrationTests):
-    """
-    Test case for the @allow_only decorator in a controller using
-    _failed_authorization() as its denial handler.
-
-    """
-
-    controller = ControllerWithAllowOnlyDecoratorAndAuthzDenialHandler
-
-    def test_authz_granted(self):
-        environ = {'REMOTE_USER': 'foobar'}
-        resp = self.app.get('/', extra_environ=environ, status=200)
-        self.assertEqual("Welcome back, foobar!", resp.body)
-
-    def test_authz_denied(self):
-        resp = self.app.get('/', status=402)
-        assert "Welcome back" not in resp.body
-
-
 class TestAllowOnlyAttributeInSubController(BaseIntegrationTests):
     """Test case for the .allow_only attribute in a sub-controller"""
 
@@ -304,37 +432,36 @@ class TestAllowOnlyAttributeInSubController(BaseIntegrationTests):
     def test_authz_granted_without_require(self):
         environ = {'REMOTE_USER': 'hiring-manager'}
         resp = self.app.get('/hr/', extra_environ=environ, status=200)
-        self.assertEqual("you can manage Human Resources", resp.body)
+        self.assertEqual("you can manage Human Resources", resp.body.decode('utf-8'))
 
     def test_authz_denied_without_require(self):
         # As an anonymous user:
         resp = self.app.get('/hr/', status=401)
-        assert "you can manage Human Resources" not in resp.body
+        assert "you can manage Human Resources" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must have been authenticated')
         # As an authenticated user:
         environ = {'REMOTE_USER': 'someone'}
         resp = self.app.get('/hr/', extra_environ = environ, status=403)
-        assert "you can manage Human Resources" not in resp.body
+        assert "you can manage Human Resources" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"hiring-manager\"')
 
     def test_authz_granted_with_require(self):
         environ = {'REMOTE_USER': 'hiring-manager'}
         resp = self.app.get('/hr/hire/gustavo', extra_environ=environ,
                             status=200)
-        self.assertEqual("gustavo was just hired", resp.body)
+        self.assertEqual("gustavo was just hired", resp.body.decode('utf-8'))
 
     def test_authz_denied_with_require(self):
         # As an anonymous user:
         resp = self.app.get('/hr/hire/gustavo', status=401)
-        assert "was just hired" not in resp.body
+        assert "was just hired" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must have been authenticated')
         # As an authenticated user:
         environ = {'REMOTE_USER': 'someone'}
         resp = self.app.get('/hr/hire/gustavo', extra_environ = environ, status=403)
-        assert "was just hired" not in resp.body
+        assert "was just hired" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"hiring-manager\"')
 
-
 class TestAllowOnlyAttributeAndDefaultAuthzDenialHandler(BaseIntegrationTests):
     """
     Test case for the .allow_only attribute in a controller using
@@ -347,12 +474,11 @@ class TestAllowOnlyAttributeAndDefaultAuthzDenialHandler(BaseIntegrationTests):
     def test_authz_granted(self):
         environ = {'REMOTE_USER': 'foobar'}
         resp = self.app.get('/', extra_environ=environ, status=200)
-        self.assertEqual("Welcome back, foobar!", resp.body)
+        self.assertEqual("Welcome back, foobar!", resp.body.decode('utf-8'))
 
     def test_authz_denied(self):
         resp = self.app.get('/', status=402)
-        assert "Welcome back" not in resp.body
-
+        assert "Welcome back" not in resp.body.decode('utf-8')
 
 class TestAppWideAuthzWithAllowOnlyDecorator(BaseIntegrationTests):
     """Test case for application-wide authz with the @allow_only decorator"""
@@ -362,22 +488,22 @@ class TestAppWideAuthzWithAllowOnlyDecorator(BaseIntegrationTests):
     def test_authz_granted_without_require(self):
         environ = {'REMOTE_USER': 'someone'}
         resp = self.app.get('/', extra_environ=environ, status=200)
-        self.assertEqual("you are in the panel", resp.body)
+        self.assertEqual("you are in the panel", resp.body.decode('utf-8'))
 
     def test_authz_denied_without_require(self):
         resp = self.app.get('/', status=401)
-        assert "you are in the panel" not in resp.body
+        assert "you are in the panel" not in resp.body.decode('utf-8')
         self._check_flash(resp, NOT_AUTHENTICATED)
 
     def test_authz_granted_with_require(self):
         environ = {'REMOTE_USER': 'admin'}
         resp = self.app.get('/add_user/foo', extra_environ=environ,
                             status=200)
-        self.assertEqual("foo was just registered", resp.body)
+        self.assertEqual("foo was just registered", resp.body.decode('utf-8'))
 
     def test_authz_denied_with_require(self):
         resp = self.app.get('/add_user/foo', status=401)
-        assert "was just registered" not in resp.body
+        assert "was just registered" not in resp.body.decode('utf-8')
         self._check_flash(resp, NOT_AUTHENTICATED)
 
 
@@ -389,34 +515,34 @@ class TestAppWideAuthzWithAllowOnlyAttribute(BaseIntegrationTests):
     def test_authz_granted_without_require(self):
         environ = {'REMOTE_USER': 'hiring-manager'}
         resp = self.app.get('/', extra_environ=environ, status=200)
-        self.assertEqual("you can manage Human Resources", resp.body)
+        self.assertEqual("you can manage Human Resources", resp.body.decode('utf-8'))
 
     def test_authz_denied_without_require(self):
         # As an anonymous user:
         resp = self.app.get('/', status=401)
-        assert "you can manage Human Resources" not in resp.body
+        assert "you can manage Human Resources" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"hiring-manager\"')
         # As an authenticated user:
         environ = {'REMOTE_USER': 'someone'}
         resp = self.app.get('/', extra_environ = environ, status=403)
-        assert "you can manage Human Resources" not in resp.body
+        assert "you can manage Human Resources" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"hiring-manager\"')
 
     def test_authz_granted_with_require(self):
         environ = {'REMOTE_USER': 'hiring-manager'}
         resp = self.app.get('/hire/gustavo', extra_environ=environ,
                             status=200)
-        self.assertEqual("gustavo was just hired", resp.body)
+        self.assertEqual("gustavo was just hired", resp.body.decode('utf-8'))
 
     def test_authz_denied_with_require(self):
         # As an anonymous user:
         resp = self.app.get('/hire/gustavo', status=401)
-        assert "was just hired" not in resp.body
+        assert "was just hired" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"hiring-manager\"')
         # As an authenticated user:
         environ = {'REMOTE_USER': 'someone'}
         resp = self.app.get('/hire/gustavo', extra_environ = environ, status=403)
-        assert "was just hired" not in resp.body
+        assert "was just hired" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"hiring-manager\"')
 
 
@@ -427,39 +553,42 @@ class TestProtectedRESTContoller(BaseIntegrationTests):
         environ = {'REMOTE_USER': 'gustavo'}
         resp = self.app.get('/rest/new', extra_environ=environ,
                             status=200)
-        self.assertEqual("new here", resp.body)
+        self.assertEqual("new here", resp.body.decode('utf-8'))
 
     def test_authz_denied(self):
         # As an anonymous user:
         resp = self.app.get('/rest/new', status=401)
-        assert "new here" not in resp.body
+        assert "new here" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"gustavo\"')
         # As an authenticated user:
         environ = {'REMOTE_USER': 'non-gustavo'}
         resp = self.app.get('/rest/new', extra_environ=environ, status=403)
-        assert "new here" not in resp.body
+        assert "new here" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"gustavo\"')
 
 
 class TestProtectedWSGIApplication(BaseIntegrationTests):
     """Test case for protected WSGI applications mounted on the controller"""
+    CONFIG_OPTIONS = {
+        'make_body_seekable': True
+    }
 
     def test_authz_granted(self):
         environ = {'REMOTE_USER': 'gustavo'}
         resp = self.app.get('/mounted_app/da-path', extra_environ=environ,
                             status=200)
-        self.assertEqual("Hello from /mounted_app/da-path", resp.body)
+        self.assertEqual("Hello from /mounted_app/da-path", resp.body.decode('utf-8'))
 
     def test_authz_denied(self):
         # As an anonymous user:
         resp = self.app.get('/mounted_app/da-path', status=401)
-        assert "Hello from /mounted_app/" not in resp.body
+        assert "Hello from /mounted_app/" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"gustavo\"')
         # As an authenticated user:
         environ = {'REMOTE_USER': 'non-gustavo'}
         resp = self.app.get('/mounted_app/da-path', extra_environ=environ,
                             status=403)
-        assert "Hello from /mounted_app/" not in resp.body
+        assert "Hello from /mounted_app/" not in resp.body.decode('utf-8')
         self._check_flash(resp, r'The current user must be \"gustavo\"')
 
 class ErrorController(object):
@@ -478,15 +607,9 @@ class TestLoggedErrorTGController(BaseIntegrationTests):
     def setUp(self):
         if not os.path.exists(session_dir):
             os.makedirs(session_dir)
+
         # Setting TG2 up:
-        c = ContextObj()
-        py_obj = PylonsContext()
-        py_obj.c = c
-        py_obj.request = py_obj.response = None
-        environ = {'pylons.routes_dict': dict(action='index'),
-                   'pylons.pylons': py_obj}
-        pylons.tmpl_context._push_object(c)
-        self.app = make_app(DefaultLessTGController, environ, with_errors=True)
+        self.app = make_app(DefaultLessTGController, {}, with_errors=True)
 
     def test_logged_index(self):
         resp = self.app.get('/index', extra_environ={'REMOTE_USER': 'gustavo'}, expect_errors=True)
