@@ -8,33 +8,38 @@ the functions they wrap, and then the DecoratedController provides the hooks
 needed to support these decorators.
 
 """
-from warnings import warn
-
-
-from decorator import decorator
-from formencode import variabledecode
-from paste.util.mimeparse import best_match
-from pylons.configuration import config
-from pylons import request, response
-from pylons.controllers.util import abort
-from webob.exc import HTTPMethodNotAllowed
-
-from tg import tmpl_context
-from tg.paginate import Page
-from tg.util import partial
-
-try:
-    from repoze.what.plugins.pylonshq import ActionProtector
-    from repoze.what.plugins.pylonshq.protectors import _BaseProtectionDecorator
-except ImportError:
-    class ActionProtector(object):
-        pass
-    class _BaseProtectionDecorator(object):
-        pass
-
-from tg.configuration import Bunch
+import copy
+import warnings
+import time
+from functools import partial
+from webob.exc import HTTPUnauthorized, HTTPMethodNotAllowed, HTTPMovedPermanently
+from tg.support import NoDefault
+from tg.support.paginate import Page
+from tg.configuration import config
+from tg.configuration.app_config import _DeprecatedControllerWrapper, call_controller
+from tg.controllers.util import abort, redirect
+from tg import tmpl_context, request, response
+from tg.util import Bunch
+from tg.configuration.sqla.balanced_session import force_request_engine
 from tg.flash import flash
-#from tg.controllers import redirect
+from tg.caching import beaker_cache, cached_property, _cached_call
+from tg.predicates import NotAuthorizedError
+from tg._compat import default_im_func, unicode_text
+from webob.acceptparse import Accept
+from tg.configuration import milestones
+import tg
+
+import logging
+log = logging.getLogger(__name__)
+
+
+def _decorated_controller_caller(tg_config, controller, remainder, params):
+    try:
+        application_controller_caller = tg_config['controller_caller']
+    except KeyError:
+        application_controller_caller = call_controller
+
+    return application_controller_caller(tg_config, controller, remainder, params)
 
 
 class Decoration(object):
@@ -42,38 +47,99 @@ class Decoration(object):
     """
     def __init__(self, controller):
         self.controller = controller
+        self.controller_caller = _decorated_controller_caller
+        self._expositions = []
         self.engines = {}
         self.engines_keys = []
+        self.default_engine = None
         self.custom_engines = {}
         self.render_custom_format = None
         self.validation = None
-        self.error_handler = None
+        self.inherit = False
+        self.requirements = []
         self.hooks = dict(before_validate=[],
                           before_call=[],
                           before_render=[],
                           after_render=[])
 
+    def __repr__(self): # pragma: no cover
+        return '<Decoration %s for %r>' % (id(self), self.controller)
+
+    @classmethod
     def get_decoration(cls, func):
-        if not hasattr(func, 'decoration'):
-            func.decoration = cls(func)
-        return func.decoration
-    get_decoration = classmethod(get_decoration)
+        try:
+            dec = func.decoration
+        except:
+            dec = func.decoration = cls(func)
+        return dec
+
+    def _register_exposition(self, exposition, inherit=False, before=False):
+        """Register an exposition for later application"""
+
+        # We need to store a reference to the exposition
+        # so that we can merge them when inheritance is performed
+        if before:
+            self._expositions.insert(0, exposition)
+        else:
+            self._expositions.append(exposition)
+
+        if inherit:
+            # if at least one exposition is in inherit mode
+            # all of them must inherit
+            self.inherit = True
+
+        milestones.renderers_ready.register(self._resolve_expositions)
+
+    def _resolve_expositions(self):
+        """Applies all the registered expositions"""
+        while True:
+            try:
+                exposition = self._expositions.pop(0)
+                exposition._apply()
+            except IndexError:
+                break
+
+    @property
+    def requirement(self):  # pragma: no cover
+        warnings.warn("Decoration.requirement is deprecated, "
+                      "please use 'requirements' instead", DeprecationWarning)
+        try:
+            return self.requirements[0]
+        except IndexError:
+            return None
 
     @property
     def exposed(self):
         return bool(self.engines) or bool(self.custom_engines)
 
-    def run_hooks(self, hook, *l, **kw):
-        for func in config.get('hooks', {}).get(hook, []):
-            func(*l, **kw)
-        for func in self.hooks[hook]:
-            func(*l, **kw)
+    def merge(self, deco):
+        # This merges already registered template engines
+        self.engines = dict(tuple(deco.engines.items()) + tuple(self.engines.items()))
+        self.engines_keys = sorted(self.engines, reverse=True)
+        self.custom_engines = dict(tuple(deco.custom_engines.items()) + tuple(self.custom_engines.items()))
 
-    def wrap_controller(self, controller):
-        controller_callable = controller
-        for wrapper in config.get('controller_wrappers', []):
-            controller_callable = wrapper(self, controller_callable)
-        return controller_callable
+        # This merges yet to register template engines
+        for exposition in reversed(deco._expositions):
+            self._register_exposition(exposition._clone(self.controller), before=True)
+
+        # inherit all the parent hooks
+        # parent hooks before current hooks so that they get called before
+        for hook_name, hooks in deco.hooks.items():
+            self.hooks[hook_name] = hooks + self.hooks[hook_name]
+
+        if not self.validation:
+            self.validation = deco.validation
+
+    def run_hooks(self, tgl, hook, *l, **kw):
+        warnings.warn("Decoration.run_hooks is deprecated, "
+                      "please use tg.hooks.notify instead", DeprecationWarning)
+        tg.hooks.notify(hook, args=l, kwargs=kw,
+                        controller=self.controller, context_config=tgl.config)
+
+    def register_hook(self, hook_name, func): #pragma: no cover
+        warnings.warn("Decoration.register_hook is deprecated, "
+                      "please use tg.hooks.register instead", DeprecationWarning)
+        tg.hooks.register(hook_name, func, controller=self.controller)
 
     def register_template_engine(self,
             content_type, engine, template, exclude_names, render_params):
@@ -95,13 +161,35 @@ class Decoration(object):
         like the rendering method or the injected doctype.
 
         """
+        default_renderer = config.get('default_renderer')
+        available_renderers = config.get('renderers', [])
 
-        self.engines[content_type or '*/*'] = (
-            engine, template, exclude_names, render_params or {})
+        if engine and not available_renderers:
+            log.warn('Renderers not registered yet while exposing template %s for engine %s, '
+                     'skipping engine availability check', template, engine)
 
-        # this is a work-around to make text/html prominent in respect
-        # to other common choices when they have the same weight for
-        # paste.util.mimeparse.best_match.
+        if engine and available_renderers and engine not in available_renderers:
+            log.debug('Registering template %s for engine %s not available. Skipping it', template, engine)
+            return
+
+        content_type = content_type or '*/*'
+        if content_type in self.engines and engine != default_renderer:
+            #Avoid overwriting the default renderer when there is already a template registered
+            return
+
+        self.engines[content_type] = (engine, template, exclude_names, render_params or {})
+
+        #Avoid engine lookup if we have only one engine registered
+        if len(self.engines) == 1:
+            self.default_engine = content_type
+        else:
+            self.default_engine = None
+
+        # This is a hack to make text/html prominent in respect to other common choices
+        # when they have the same weight for webob.acceptparse.Accept.best_match().
+        # It uses the fact that the most common content types are all alphabetically
+        # precedent to text/html, and so sorting engine keys alphabetically reversed
+        # should make text/html the first choice when no other better choices are available.
         self.engines_keys = sorted(self.engines, reverse=True)
 
     def register_custom_template_engine(self, custom_format,
@@ -131,7 +219,7 @@ class Decoration(object):
         self.custom_engines[custom_format or '"*/*"'] = (
             content_type, engine, template, exclude_names, render_params or {})
 
-    def lookup_template_engine(self, request):
+    def lookup_template_engine(self, tgl):
         """Return the template engine data.
 
         Provides a convenience method to get the proper engine,
@@ -139,16 +227,11 @@ class Decoration(object):
         tg_format (which is pulled off of the request headers).
 
         """
-
-        if hasattr(request, 'response_type'
-                ) and request.response_type in self.engines:
-            accept_types = request.response_type
-        else:
-            accept_types = request.headers.get('accept', '*/*')
+        request = tgl.request
+        response = tgl.response
 
         try:
-            render_custom_format = request._render_custom_format[
-                self.controller]
+            render_custom_format = request._render_custom_format[self.controller]
         except:
             render_custom_format = self.render_custom_format
 
@@ -156,27 +239,29 @@ class Decoration(object):
             (content_type, engine, template, exclude_names, render_params
                 ) = self.custom_engines[render_custom_format]
         else:
-            if self.engines:
-                content_type = best_match(self.engines_keys, accept_types)
+            if self.default_engine:
+                content_type = self.default_engine
+            elif self.engines:
+                if request._response_type and request._response_type in self.engines:
+                    accept_types = request._response_type
+                else:
+                    accept_types = request.headers.get('accept', '*/*')
+                content_type = Accept(accept_types).best_match(self.engines_keys, self.engines_keys[0])
             else:
                 content_type = 'text/html'
 
-            if content_type == 'CUSTOM/LEAVE':
-                warn('@expose(CUSTOM_CONTENT_TYPE) is no longer needed'
-                     ' and should be replaced with @expose()')
-
             # check for overridden content type from the controller call
-            controller_content_type = response.headers.get('Content-Type')
-
-            if controller_content_type:
-                # make sure we handle types like 'text/html; charset=utf-8'
-                content_type = controller_content_type.split(';', 1)[0]
+            try:
+                controller_content_type = response.headers['Content-Type']
+                # make sure we handle content_types like 'text/html; charset=utf-8'
+                content_type = controller_content_type.split(';')[0]
+            except KeyError:
+                pass
 
             # check for overridden templates
             try:
-                (engine, template, exclude_names, render_params
-                    ) = request._override_mapping[
-                            self.controller][content_type.split(';', 1)[0]]
+                cnt_override_mapping = request._override_mapping[self.controller]
+                engine, template, exclude_names, render_params = cnt_override_mapping[content_type.split(";")[0]]
             except (AttributeError, KeyError):
                 (engine, template, exclude_names, render_params
                     ) = self.engines.get(content_type, (None,) * 4)
@@ -184,20 +269,34 @@ class Decoration(object):
         if 'charset' not in content_type and (content_type.startswith('text')
                 or  content_type in ('application/xhtml+xml',
                         'application/xml', 'application/json')):
-            content_type = '%s; charset=utf-8' % content_type
+            content_type += '; charset=utf-8'
 
         return content_type, engine, template, exclude_names, render_params
 
-    def register_hook(self, hook_name, func):
+    def _register_hook(self, hook_name, func):
         """Registers the specified function as a hook.
 
+        This is internal API which is used by tg.hooks, instead of
+        calling this tg.hooks.register should be used.
+
         We now have four core hooks that can be applied by adding
         decorators: before_validate, before_call, before_render, and
         after_render. register_hook attaches the function to the hook
         which get's called at the appropriate time in the request life
         cycle.)
         """
-        self.hooks[hook_name].append(func)
+        self.hooks.setdefault(hook_name, []).append(func)
+
+    def _register_requirement(self, requirement):
+        self._register_hook('before_call', requirement._check_authorization)
+        self.requirements.append(requirement)
+
+    def _register_controller_wrapper(self, wrapper):
+        try:
+            self.controller_caller = wrapper(self.controller_caller)
+        except TypeError:
+            self.controller_caller = _DeprecatedControllerWrapper(wrapper, tg.config,
+                                                                  self.controller_caller)
 
 
 class _hook_decorator(object):
@@ -210,11 +309,15 @@ class _hook_decorator(object):
     hook_name = None
 
     def __init__(self, hook_func):
+        if hasattr(hook_func, '__name__'):
+            self.__name__ = hook_func.__name__
+        if hasattr(hook_func, '__doc__'):
+            self.__doc__ = hook_func.__doc__
         self.hook_func = hook_func
 
     def __call__(self, func):
         deco = Decoration.get_decoration(func)
-        deco.register_hook(self.hook_name, self.hook_func)
+        deco._register_hook(self.hook_name, self.hook_func)
         return func
 
 
@@ -259,8 +362,16 @@ class expose(object):
         The default content type is 'text/html'.
       exclude_names
         Assign exclude names
+      custom_format
+        Registers as a custom format which can later be activated calling
+        use_custom_format
       render_params
         Assign parameters that shall be passed to the rendering method.
+      inherit
+        Inherit all the decorations from the same method in the parent
+        class. This will let the exposed method expose the same template
+        as the overridden method template and keep the same hooks and
+        validation that the parent method had.
 
     The expose decorator registers a number of attributes on the
     decorated function, but does not actually wrap the function the way
@@ -288,7 +399,7 @@ class expose(object):
         def my_exposed_method(self):
             return dict(a=1, b=2, d="username")
 
-    The expose('json') syntax is a special case.  json is a buffet
+    The expose('json') syntax is a special case.  json is a
     rendering engine, but unlike others it does not require a template,
     and expose assumes that it matches content_type='application/json'
 
@@ -315,45 +426,83 @@ class expose(object):
     """
 
     def __init__(self, template='', content_type=None, exclude_names=None,
-                 custom_format=None, render_params=None):
-        if exclude_names is None:
-            exclude_names = []
+                 custom_format=None, render_params=None, inherit=False):
 
-        if template in config.get('renderers', []):
-            engine, template = template, ''
-
-        elif ':' in template:
-            engine, template = template.split(':', 1)
-
-        elif template:
-            # Use the default templating engine from the config
-            if config.get('use_legacy_renderer'):
-                engine = config['buffet.template_engines'][0]['engine']
-
-            else:
-                engine = config.get('default_renderer')
-
-        else:
-            engine, template = None, None
-
-        if content_type is None:
-            if engine == 'json':
-                content_type = 'application/json'
-            else:
-                content_type = 'text/html'
-
-        if engine in ('json', 'amf') and 'tmpl_context' not in exclude_names:
-            exclude_names.append('tmpl_context')
-
-        self.engine = engine
+        self.engine = None
         self.template = template
         self.content_type = content_type
         self.exclude_names = exclude_names
         self.custom_format = custom_format
         self.render_params = render_params
 
+        self.inherit = inherit
+        self._func = None
+
     def __call__(self, func):
+        self._func = func
         deco = Decoration.get_decoration(func)
+        deco._register_exposition(self, self.inherit)
+        return func
+
+    def _resolve_options(self):
+        """This resolves exposition options that depend on
+        configuration steps that might not have already happened.
+        It's automatically called by _apply when required
+
+        """
+        if self.engine is not None:
+            return
+
+        exclude_names = self.exclude_names
+        template = self.template
+        content_type = self.content_type
+
+        if exclude_names is None:
+            exclude_names = []
+
+        if template in config.get('renderers', []):
+            engine, template = template, ''
+        elif ':' in template:
+            engine, template = template.split(':', 1)
+        elif template:
+            # Use the default templating engine from the config
+            engine = config.get('default_renderer')
+        else:
+            engine, template = None, None
+
+        if content_type is None:
+            all_engines_options = config.get('rendering_engines_options', {})
+            engine_options = all_engines_options.get(engine, {})
+            content_type = engine_options.get('content_type', 'text/html')
+
+        engines_without_vars = config.get('rendering_engines_without_vars', [])
+        if engine in engines_without_vars and 'tmpl_context' not in exclude_names:
+            exclude_names.append('tmpl_context')
+
+        self.engine = engine
+        self.template = template
+        self.content_type = content_type
+        self.exclude_names = exclude_names
+
+    def _clone(self, func):
+        clone = copy.copy(self)
+        clone._func = func
+        return clone
+
+    def _apply(self):
+        """Applies an exposition for real"""
+        if self._func is None:
+            log.error('Applying an exposition with no decorated function!')
+            return
+
+        self._resolve_options()
+
+        deco = Decoration.get_decoration(self._func)
+        if deco.inherit and not self.template and not self.engine:
+            # If we are just inheriting without adding additional
+            # engines or templates we can just skip this part.
+            return
+
         if self.custom_format:
             deco.register_custom_template_engine(
                 self.custom_format, self.content_type, self.engine,
@@ -362,7 +511,6 @@ class expose(object):
             deco.register_template_engine(
                 self.content_type, self.engine,
                 self.template, self.exclude_names, self.render_params)
-        return func
 
 
 def use_custom_format(controller, custom_format):
@@ -378,16 +526,19 @@ def use_custom_format(controller, custom_format):
         render_custom_format = request._render_custom_format
     except AttributeError:
         render_custom_format = request._render_custom_format = {}
-    render_custom_format[controller.im_func] = custom_format
+    render_custom_format[default_im_func(controller)] = custom_format
 
 
-def override_template(controller, template):
+def override_template(view, template):
     """Override the template to be used.
 
-    Use override_template in a controller in order to change the template
+    Use override_template in a controller method in order to change the template
     that will be used to render the response dictionary dynamically.
 
-    The template string passed in requires that
+    The ``view`` argument is the actual controller method for which you
+    want to replace the template.
+
+    The ``template`` string passed in requires that
     you include the template engine name, even if you're using the default.
 
     So you have to pass in a template id string like::
@@ -399,18 +550,18 @@ def override_template(controller, template):
 
     """
     try:
-        engines = controller.decoration.engines
+        engines = view.decoration.engines
     except:
         return
 
-    for content_type, content_engine in engines.iteritems():
+    for content_type, content_engine in engines.items():
         tmpl = template.split(':', 1)
         tmpl.extend(content_engine[2:])
         try:
             override_mapping = request._override_mapping
         except AttributeError:
             override_mapping = request._override_mapping = {}
-        override_mapping.setdefault(controller.im_func, {}).update({content_type: tmpl})
+        override_mapping.setdefault(default_im_func(view), {}).update({content_type: tmpl})
 
 
 class validate(object):
@@ -436,6 +587,7 @@ class validate(object):
 
     """
     def __init__(self, validators=None, error_handler=None, form=None):
+        self.validators = None
         if form:
             self.validators = form
         if validators:
@@ -448,6 +600,29 @@ class validate(object):
         return func
 
 
+class decode_params(object):
+    """Decorator that enables parsing parameters from request body.
+
+    By default the arguments are parsed in **JSON** format (which is
+    currently the only supported format).
+
+    """
+    def __init__(self, format='json'):
+        if format not in ('json', ):
+            raise ValueError('Currently only JSON format is supported')
+
+        self._format = format
+
+    def run_hook(self, remainder, params):
+        if self._format == 'json' and request.content_type == 'application/json':
+            params.update(request.json_body)
+
+    def __call__(self, func):
+        decoration = Decoration.get_decoration(func)
+        decoration._register_hook('before_validate', self.run_hook)
+        return func
+
+
 class paginate(object):
     """Paginate a given collection.
 
@@ -501,8 +676,8 @@ class paginate(object):
 
     def __call__(self, func):
         decoration = Decoration.get_decoration(func)
-        decoration.register_hook('before_validate', self.before_validate)
-        decoration.register_hook('before_render', self.before_render)
+        decoration._register_hook('before_validate', self.before_validate)
+        decoration._register_hook('before_render', self.before_render)
         return func
 
     def before_validate(self, remainder, params):
@@ -549,8 +724,7 @@ class paginate(object):
 
         paginator = request.paginators[self.name]
         collection = output[self.name]
-        page = Page(collection, paginator.paginate_page,
-            paginator.paginate_items_per_page, controller='/')
+        page = Page(collection, paginator.paginate_page, paginator.paginate_items_per_page)
         page.kwargs = paginator.paginate_params
         if self.page_param != 'name':
             page.pager = partial(page.pager, page_param=self.page_param)
@@ -558,34 +732,16 @@ class paginate(object):
             tmpl_context.paginators = Bunch()
         tmpl_context.paginators[self.name] = output[self.name] = page
 
-
-@decorator
-def postpone_commits(func, *args, **kwargs):
-    """Turns SQLAlchemy commits into flushes in the decorated method.
-
-    This has the end-result of postponing the commit to the normal TG2
-    transaction boundary. """
-
-    #TODO: Test and document this.
-    s = config.get('DBSession', None)
-    assert hasattr(s, 'commit')
-    old_commit = s.commit
-    s.commit = s.flush
-    retval = func(*args, **kwargs)
-    s.commit = old_commit
-    return retval
-
-
 @before_validate
 def https(remainder, params):
     """Ensure that the decorated method is always called with https."""
-    from tg.controllers import redirect
     if request.scheme.lower() == 'https': return
     if request.method.upper() == 'GET':
         redirect('https' + request.url[len(request.scheme):])
-    raise HTTPMethodNotAllowed(headers=dict(Allow='GET')).exception
+    raise HTTPMethodNotAllowed(headers=dict(Allow='GET'))
 
 
+_variabledecode = None
 @before_validate
 def variable_decode(remainder, params):
     """Best-effort formencode.variabledecode on the params before validation.
@@ -597,8 +753,12 @@ def variable_decode(remainder, params):
     to {'foo-1':'1', 'foo-2':'2', 'foo':['1', '2']}.
 
     """
+    global _variabledecode
+    if _variabledecode is None:
+        from formencode import variabledecode as _variabledecode
+
     try:
-        new_params = variabledecode.variable_decode(params)
+        new_params = _variabledecode.variable_decode(params)
         params.update(new_params)
     except:
         pass
@@ -625,9 +785,9 @@ def without_trailing_slash(remainder, params):
     In addition, the URL http://localhost:8080/sample/1/ redirects to http://localhost:8080/sample/1
 
     """
-    if request.method == 'GET' and request.path.endswith('/') and not(request.response_type) and len(request.params)==0:
-        from tg.controllers import redirect
-        redirect(request.url[:-1])
+    req = request._current_obj()
+    if req.method == 'GET' and req.path.endswith('/') and not(req._response_type) and len(req.params)==0:
+        redirect(request.url[:-1], redirect_with=HTTPMovedPermanently)
 
 
 @before_validate
@@ -651,79 +811,202 @@ def with_trailing_slash(remainder, params):
     In addition, the URL http://localhost:8080/sample/1 redirects to http://localhost:8080/sample/1/
 
     """
-    if (request.method == 'GET'
-        and not(request.path.endswith('/'))
-        and not(request.response_type)
-        and len(request.params)==0):
-        from tg.controllers import redirect
-        redirect(request.url+'/')
+    req = request._current_obj()
+    if (req.method == 'GET' and not(req.path.endswith('/')) and not(req._response_type) and len(req.params)==0):
+        redirect(request.url+'/', redirect_with=HTTPMovedPermanently)
 
 
-#{ Authorization decorators
-class require(ActionProtector):
+class require(object):
     """
-    TurboGears-specific repoze.what-pylons action protector.
+    Decorator that checks if the specified predicate it met, if it isn't
+    it calls the denial_handler to prevent access to the decorated method.
 
     The default authorization denial handler of this protector will flash
     the message of the unmet predicate with ``warning`` or ``error`` as the
     flash status if the HTTP status code is 401 or 403, respectively.
 
-    See :class:`allow_only` for controller-wide authorization.
+    :param predicate: An object with a check_authorization(environ) method which
+        must raise a tg.predicates.NotAuthorizedError if not met.
+    :param denial_handler: The callable to be run if authorization is
+        denied (overrides :attr:`default_denial_handler` if defined).
+    :param smart_denial: A list of response types for which to trigger
+        the smart denial, which will act as an API providing a pass-through
+        :func:`tg.controllers.util.abort`.
+        If ``True``, ``('application/json', 'text/xml')`` will be used.
+
+    If called, ``denial_handler`` will be passed a positional argument
+    which represents a message on why authorization was denied.
+
+    Use ``allow_only`` property of ``TGController`` for controller-wide authorization.
 
     """
-    def default_denial_handler(self, reason):
-        """Authorization denial handler for repoze.what-pylons protectors."""
-        if response.status_int == 401:
-            status = 'warning'
-        else:
-            # Status is a 403
-            status = 'error'
-        flash(reason, status=status)
-        abort(response.status_int, reason)
+    def __init__(self, predicate, denial_handler=None, smart_denial=False):
+        self.predicate = predicate
+        self.denial_handler = denial_handler or self.default_denial_handler
 
+        if smart_denial is True:
+            smart_denial = ('application/json', 'text/xml')
+        self.smart_denial = smart_denial
 
-class allow_only(_BaseProtectionDecorator):
-    """
-    TurboGears-specific repoze.what-pylons controller protector.
+    def __call__(self, func):
+        deco = Decoration.get_decoration(func)
+        deco._register_requirement(self)
+        return func
 
-    The default authorization denial handler of this protector will flash
-    the message of the unmet predicate with ``warning`` or ``error`` as the
-    flash status if the HTTP status code is 401 or 403, respectively, since
-    by default the ``__before__`` method of the controller is decorated with
-    :class:`require`.
+    def _check_authorization(self, *args, **kwargs):
+        req = request._current_obj()
 
-    If the controller class has the ``_failed_authorization`` *class method*,
-    it will replace the default denial handler.
-
-    """
-    protector = require
-
-    def __call__(self, cls, *args, **kwargs):
-        if hasattr(self.protector, 'predicate'):
-            cls.allow_only = self.protector.predicate
-        if hasattr(cls, '_failed_authorization'):
-            self.denial_handler = cls._failed_authorization
-        sup = super(allow_only, self)
-        if hasattr(sup, '__call__'):
-            return super(allow_only, self).__call__(cls, *args, **kwargs)
-
-
-class cached_property(object):
-    """Cached property executing the getter only once."""
-
-    def __init__(self, func):
-        self.__name__ = func.__name__
-        self.__module__ = func.__module__
-        self.__doc__ = func.__doc__
-        self.func = func
-
-    def __get__(self, obj, type=None):
-        if obj is None:
-            return self
         try:
-            value = obj.__dict__[self.__name__]
-        except KeyError:
-            value = obj.__dict__[self.__name__] = self.func(obj)
-        return value
+            self.predicate.check_authorization(req.environ)
+        except NotAuthorizedError as e:
+            reason = unicode_text(e)
+            if req.environ.get('repoze.who.identity'):
+                # The user is authenticated.
+                code = 403
+            else:
+                # The user is not authenticated.
+                code = 401
+            response.status = code
+            return self.denial_handler(reason)
 
-#}
+    def default_denial_handler(self, reason):
+        """Authorization denial handler for protectors."""
+        passthrough_abort = False
+
+        if self.smart_denial:
+            response_type = response.content_type or request.response_type
+            if response_type in self.smart_denial:
+                # It's an API response, use a pass-through abort
+                passthrough_abort = True
+                if response_type == 'application/json':
+                    passthrough_abort = 'json'
+
+        if passthrough_abort is False:
+            # Plain HTML page
+            status = 'warning' if response.status_int == 401 else 'error'
+            flash(reason, status=status)
+
+        abort(response.status_int, reason, passthrough=passthrough_abort)
+
+
+class with_engine(object):
+    """
+    Decorator to force usage of a specific database engine
+    in TurboGears SQLAlchemy BalancedSession.
+
+    :param engine_name: 'master' or the name of one of the slaves, if is ``None``
+             it will not force any specific engine.
+    :param master_params: A dictionary or GET parameters that when present will force
+             usage of the master node. The keys of the dictionary will be the
+             name of the parameters to look for, while the values must be whenever
+             to pop the paramter from the parameters passed to the controller (True/False).
+             If `master_params` is a list then it is converted to a dictionary where
+             the keys are the entries of the list and the value is always True.
+    """
+
+    def __init__(self, engine_name=None, master_params={}):
+        self.engine_name = engine_name
+
+        if not hasattr(master_params, 'keys'):
+            self.master_params = dict((p, True) for p in master_params)
+        else:
+            self.master_params = master_params
+
+    def before_validate(self, remainder, params):
+        force_request_engine(self.engine_name)
+        for p, pop in self.master_params.items():
+            if p in params:
+                if pop:
+                    v = params.pop(p, None)
+                else:
+                    v = params.get(p)
+
+                if v:
+                    force_request_engine('master')
+                    break
+
+    def __call__(self, func):
+        decoration = Decoration.get_decoration(func)
+        decoration._register_hook('before_validate', self.before_validate)
+        return func
+
+
+class cached(object):
+    """
+    Decorator to cache the controller, if you also want to cache
+    template remember to return ``tg_cache`` option from the controller.
+
+    The following parameters are accepted:
+
+    ``key`` - Specifies the controller parameters used to generate the cache key.
+        NoDefault - Uses function name and all request parameters as the key (default)
+
+        None - No variable key, uses only function name as key
+
+        string - Use function name and only "key" parameter
+
+        list - Use function name and all parameters listed
+    ``expire``
+        Time in seconds before cache expires, or the string "never".
+        Defaults to "never"
+    ``type``
+        Type of cache to use: dbm, memory, file, memcached, or None for
+        Beaker's default
+    ``cache_headers``
+        A tuple of header names indicating response headers that
+        will also be cached.
+    ``invalidate_on_startup``
+        If True, the cache will be invalidated each time the application
+        starts or is restarted.
+    ``cache_response``
+        Determines whether the response at the time the cache is used
+        should be cached or not, defaults to True.
+
+        .. note::
+            When cache_response is set to False, the cache_headers
+            argument is ignored as none of the response is cached.
+    """
+    def __init__(self, key=NoDefault, expire="never", type=None,
+                 query_args=None,  # Backward compatibility, actually ignored
+                 cache_headers=('content-type', 'content-length'),
+                 invalidate_on_startup=False, cache_response=True,
+                 **b_kwargs):
+        self.key = key
+        self.expire = expire
+        self.type = type
+        self.cache_headers = cache_headers
+        self.invalidate_on_startup = invalidate_on_startup
+        self.cache_response = cache_response
+        self.beaker_options = b_kwargs
+
+    def __call__(self, func):
+        decoration = Decoration.get_decoration(func)
+
+        def controller_wrapper(__, next_caller):
+            if self.invalidate_on_startup:
+                starttime = time.time()
+            else:
+                starttime = None
+
+            def cached_call_controller(controller, remainder, params):
+                if self.key:
+                    key_dict = request.args_params
+                    if self.key != NoDefault:
+                        if isinstance(self.key, (list, tuple)):
+                            key_dict = dict((k, key_dict[k]) for k in key_dict)
+                        else:
+                            key_dict = {self.key: key_dict[self.key]}
+                else:
+                    key_dict = {}
+
+                return _cached_call(next_caller, (controller, remainder, params), {},
+                                    key_func=func, key_dict=key_dict,
+                                    expire=self.expire, type=self.type,
+                                    starttime=starttime, cache_headers=self.cache_headers,
+                                    cache_response=self.cache_response,
+                                    cache_extra_args=self.beaker_options)
+
+            return cached_call_controller
+
+        decoration._register_controller_wrapper(controller_wrapper)
+        return func
